'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.default = function (_ref) {
  var t = _ref.types;

  function isNewlyCreatedPath(path) {
    return !path.node.loc;
  }

  function getOffsettedRange(start, end, offset) {
    return {
      start: start - offset,
      end: end - offset
    };
  }

  function getBindingsForScope(scope, allBindings) {
    var bindings = [];
    var currScope = scope;

    while (currScope) {
      var scopeBindings = allBindings[currScope.uid];
      if (scopeBindings) {
        bindings.unshift.apply(bindings, _toConsumableArray(scopeBindings));
      }
      currScope = currScope.parent;
    }

    return bindings;
  }

  function createTraceCall(_ref2) {
    var start = _ref2.start,
        end = _ref2.end,
        bindings = _ref2.bindings,
        compared = _ref2.compared,
        returnValue = _ref2.returnValue,
        enterCall = _ref2.enterCall,
        leaveCall = _ref2.leaveCall;

    var stepProps = [t.objectProperty(t.identifier('highlight'), t.objectExpression([t.objectProperty(t.identifier('start'), t.numericLiteral(start)), t.objectProperty(t.identifier('end'), t.numericLiteral(end))])), t.objectProperty(t.identifier('bindings'), t.objectExpression(bindings.map(function (name) {
      return t.objectProperty(t.identifier(name), t.callExpression(t.identifier('cloneDeep'), [t.identifier(name)]));
    })))];
    if (compared !== undefined) {
      stepProps.push(t.objectProperty(t.identifier('compared'), t.arrayExpression(compared.map(function (s) {
        return t.stringLiteral(s);
      }))));
    }
    if (returnValue !== undefined) {
      stepProps.push(t.objectProperty(t.identifier('returnValue'), returnValue));
    }

    var args = [t.objectProperty(t.identifier('step'), t.objectExpression(stepProps))];
    if (enterCall) {
      args.push(t.objectProperty(t.identifier('enterCall'), t.booleanLiteral(true)));
    }
    if (leaveCall) {
      args.push(t.objectProperty(t.identifier('leaveCall'), t.booleanLiteral(true)));
    }

    return t.expressionStatement(t.callExpression(t.identifier('__trace'), [t.objectExpression(args)]));
  }

  var innerVisitor = {
    /**
     * Prepend trace call(enter) BEFORE anything in the main function's body.
     * Highlight contains function signature "fn(args)".
     */
    BlockStatement: function BlockStatement(path) {
      var parentPath = path.parentPath;
      var parentNode = parentPath.node;
      if (!parentNode.id || parentNode.id.name !== this.fnName) {
        return;
      }

      var start = parentPath.get('id').node.start;
      var end = parentNode.params[parentNode.params.length - 1].end + 1;
      path.unshiftContainer('body', createTraceCall(_extends({
        bindings: getBindingsForScope(path.scope, this.bindings)
      }, getOffsettedRange(start, end, this.fnOffset), {
        enterCall: true
      })));
    },


    /**
     * Append trace call() AFTER var declaration. Highlight contains entire
     * declaration.
     *
     * Created vars are pushed to visitor state to be included in bindings of
     * this and future trace() calls
     */
    VariableDeclaration: function VariableDeclaration(path) {
      var _scopeBindings;

      // Ignore var declarations we create
      if (isNewlyCreatedPath(path)) {
        return;
      }

      var scopeId = path.scope.uid;
      var scopeBindings = this.bindings[scopeId];
      if (!scopeBindings) {
        scopeBindings = this.bindings[scopeId] = [];
      }
      (_scopeBindings = scopeBindings).push.apply(_scopeBindings, _toConsumableArray(path.node.declarations.map(function (d) {
        return d.id.name;
      })));

      var _path$node = path.node,
          start = _path$node.start,
          end = _path$node.end;

      path.insertAfter(createTraceCall(_extends({
        bindings: getBindingsForScope(path.scope, this.bindings)
      }, getOffsettedRange(start, end, this.fnOffset))));
    },


    /**
     * Append trace call() AFTER assignment expressions. Highlight contains
     * entire expression. Bindings include new values.
     */
    AssignmentExpression: function AssignmentExpression(path) {
      var _path$node2 = path.node,
          start = _path$node2.start,
          end = _path$node2.end;

      path.insertAfter(createTraceCall(_extends({
        bindings: getBindingsForScope(path.scope, this.bindings)
      }, getOffsettedRange(start, end, this.fnOffset))));
    },


    /**
     * Replace test expressions with IIFE containing trace() call BEFORE
     * test value has been computed. Highlight contains test expression only.
     */
    'WhileStatement|IfStatement': function WhileStatementIfStatement(path) {
      var testPath = path.get('test');
      var _testPath$node = testPath.node,
          start = _testPath$node.start,
          end = _testPath$node.end,
          left = _testPath$node.left,
          right = _testPath$node.right,
          type = _testPath$node.type;
      var code = path.hub.file.code;

      var compared = type === 'BinaryExpression' ? [code.slice(left.start, left.end), code.slice(right.start, right.end)] : [code.slice(start, end)];
      testPath.replaceWith(t.callExpression(t.arrowFunctionExpression([], t.blockStatement([createTraceCall(_extends({
        bindings: getBindingsForScope(path.scope, this.bindings)
      }, getOffsettedRange(start, end, this.fnOffset), {
        compared: compared
      })), t.returnStatement(testPath.node)])), []));
    },
    CallExpression: function CallExpression(path) {
      // This prevents an infinite loop
      if (this.processedCallExpressions.indexOf(path.node) !== -1) {
        return;
      }

      this.processedCallExpressions.push(path.node);

      var node = path.node;
      var start = node.start,
          end = node.end;

      var bindings = getBindingsForScope(path.scope, this.bindings);
      var traceCall = createTraceCall(_extends({
        bindings: bindings
      }, getOffsettedRange(start, end, this.fnOffset)));

      /**
       * Replace recursive call expressions with IIFE containing trace() call
       * BEFORE return value has been computed.
       */
      if (node.callee.name === this.fnName) {
        path.replaceWith(t.callExpression(t.arrowFunctionExpression([], t.blockStatement([traceCall, t.returnStatement(node)])), []));
      }

      /**
       * Replace mutative call expressions with IIFE containing trace() call
       * AFTER object has been mutated.
       */
      if (node.callee.type === 'MemberExpression' && bindings.indexOf(node.callee.object.name) !== -1 && MUTATOR_METHODS.indexOf(node.callee.property.name) !== -1) {
        var returnValId = path.scope.generateUidIdentifier('uid');
        path.replaceWith(t.callExpression(t.arrowFunctionExpression([], t.blockStatement([t.variableDeclaration('const', [t.variableDeclarator(returnValId, node)]), traceCall, t.returnStatement(returnValId)])), []));
      }
    },


    /**
     * Replace return calls with IIFE containing trace(leave) call AFTER return
     * value has been computed, before it is returned. Highlight contains
     * entire return statement.
     */
    ReturnStatement: function ReturnStatement(path) {
      // Ignore return statements we create
      if (isNewlyCreatedPath(path)) {
        return;
      }

      var returnValId = path.scope.generateUidIdentifier('uid');
      var argumentPath = path.get('argument');
      var _path$node3 = path.node,
          start = _path$node3.start,
          end = _path$node3.end;


      argumentPath.replaceWith(t.callExpression(t.arrowFunctionExpression([], t.blockStatement([t.variableDeclaration('const', [t.variableDeclarator(returnValId, argumentPath.node)]), createTraceCall(_extends({
        bindings: getBindingsForScope(path.scope, this.bindings)
      }, getOffsettedRange(start, end, this.fnOffset), {
        leaveCall: true,
        returnValue: returnValId
      })), t.returnStatement(returnValId)])), []));
    }
  };

  return {
    visitor: {
      ExportDefaultDeclaration: function ExportDefaultDeclaration(path) {
        var fnPath = path.get('declaration');
        var fnName = fnPath.node.id.name;
        var _fnPath$node = fnPath.node,
            start = _fnPath$node.start,
            end = _fnPath$node.end;


        path.replaceWithMultiple(buildTemplate({
          ALGORITHM_NAME: t.identifier(fnName),
          ALGORITHM_BODY: fnPath.node,
          ALGORITHM_CODE: t.stringLiteral(path.hub.file.code.slice(start, end))
        }));

        var bodyPath = fnPath.get('body');
        var params = Object.keys(bodyPath.scope.getAllBindingsOfKind('param'));

        fnPath.traverse(innerVisitor, {
          fnName: fnName,
          bindings: _defineProperty({}, fnPath.scope.uid, params),
          processedCallExpressions: [],
          fnOffset: start
        });
      }
    }
  };
};

var _babelTemplate = require('babel-template');

var _babelTemplate2 = _interopRequireDefault(_babelTemplate);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var buildTemplate = (0, _babelTemplate2.default)('\nconst cloneDeep = require(\'lodash.clonedeep\');\n\nlet __steps;\nlet __parentStepId;\nlet __lastStepId;\n\nfunction __enterCall() {\n  // Mark steps from sub call as decendends of last step from curr call\n  __parentStepId = __lastStepId;\n}\n\nfunction __leaveCall() {\n  if (__parentStepId) {\n    // Revert step refs in parent callee\n    __lastStepId = __parentStepId;\n    __parentStepId = __steps[__parentStepId].parentStepId;\n  }\n}\n\nfunction __trace({\n  step,\n  enterCall,\n  leaveCall,\n}) {\n  if (enterCall) {\n    __enterCall();\n  }\n\n  __steps.push({\n    parentStepId: __parentStepId,\n    ...step,\n  });\n\n  // Track last step id to reference back to entering a subcall\n  __lastStepId = __steps.length - 1;\n\n  if (leaveCall) {\n    __leaveCall();\n  }\n}\n\nALGORITHM_BODY\n\nfunction __traceWrapper(...args) {\n  __steps = [];\n  __parentStepId = undefined;\n  __lastStepId = undefined;\n\n  const returnValue = ALGORITHM_NAME(...args);\n  return {\n    steps: __steps,\n    returnValue: returnValue,\n  };\n};\n\n__traceWrapper.code = ALGORITHM_CODE;\n\nmodule.exports = __traceWrapper;\n', {
  plugins: ['objectRestSpread']
});

var MUTATOR_METHODS = ['push', 'shift'];